% (C) - Boris Bellalta - 2021


function ExampleLecture7(seed)

rng(seed);

C1 = 10E6;
C2 = 10E6;
L=8000;
KS=100;
B=8E6;

% Sim parameters
TimeHorizon=400; % Increase to 400 for the non-stationary case
TimeChange=200;

N=2; % number of paths

a1=zeros(1,TimeHorizon);
a1(1:TimeChange)=0.1.*ones(1,TimeChange);
a1(TimeChange+1:TimeHorizon)=0.3.*ones(1,TimeHorizon-TimeChange);

b1=zeros(1,TimeHorizon);
b1(1:TimeChange)=0.2.*ones(1,TimeChange);
b1(TimeChange+1:TimeHorizon)=0.6.*ones(1,TimeHorizon-TimeChange);

a2=0.2;
b2=0.4;

w(1,:)=a1 + (b1-a1).*rand(1,TimeHorizon);
w(2,:)=a2 + (b2-a2).*rand(1,TimeHorizon);

% Weights - Actions
%alfa = [0:0.05:1];
%alfa = [0 0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1];
alfa = [0 0.1 0.25 0.5 0.75 0.9 1];
%alfa = [0.5];


K = length(alfa);

% Gamma
gamma = 0.9;

%----------------- Epsilon greedy ------------------

epsilon = 0.1;
Delta = 0.0;

%epsilon = 1;
%Delta = 0.01;

Q_EG_disp = zeros(K,TimeHorizon);
Q_EG = zeros(K,1);

R_EG = zeros(1,TimeHorizon);
alfa_EG = zeros(1,TimeHorizon);
alfa_selected_EG = zeros(1,K);

% First Iteration

n_sel = randi(K,1); % Action Selected, random
alfa_EG(1) = n_sel;
alfa_selected_EG(n_sel) = 1;

% First Path
[ED1(1),ELosses1(1)] = PathDelay(KS,alfa(n_sel)*B,L,C1,w(1,1));
% Second Path
[ED2(1),ELosses2(1)] = PathDelay(KS,(1-alfa(n_sel))*B,L,C2,w(1,1));

% Reward (taking only the delay into account)
R_EG(1) = alfa(n_sel)*ED1(1)+(1-alfa(n_sel))*ED2(1); % Mean
%R_EG(1) = max([ED1(1) ED2(1)]); % Max

% Update Q-table
Q_EG(n_sel)=R_EG(1);
Q_EG_Disp(:,1) = Q_EG;

alfa_selected_EG(n_sel) = 1;

for t=2:TimeHorizon
   
    if(rand < epsilon) % explore
        n_sel=randi(K,1);
    else % exploit
        [~,n_sel]=min(Q_EG(:)); % Maximize the reward --> minimize the delay!!!
    end
    
    alfa_EG(t) = n_sel;
    alfa_selected_EG(n_sel)=alfa_selected_EG(n_sel)+1;
    
    % First Path
    [ED1(t),ELosses1(t)] = PathDelay(KS,alfa(n_sel)*B,L,C1,w(1,t));
    % Second Path
    [ED2(t),ELosses2(t)] = PathDelay(KS,(1-alfa(n_sel))*B,L,C2,w(2,t));

    % ----------Rewards---------------------
    %R_EG(t) = alfa(n_sel)*ED1(t)+(1-alfa(n_sel))*ED2(t); % Average Delay
    R_EG(t) = abs(ED1(t)-ED2(t)); % Diference between delays
    %if(t<10)
    %    R_EG(t) = max([ED1(1:t) ED2(1:t)]); % Max delay
    %else
    %    R_EG(t) = max([ED1(t-9:t) ED2(t-9:t)]); % Max delay
    %end
    
    % --------- Q table ---------------------
    %Q_EG(n_sel)=(Q_EG(n_sel)*(alfa_selected_EG(n_sel)-1) + R_EG(t))/alfa_selected_EG(n_sel); % Average    
    Q_EG(n_sel)=gamma*R_EG(t)+(1-gamma)*Q_EG(n_sel); % Gamma method
    
    Q_EG_disp(:,t)=Q_EG(:); % For display purposes
        
    % Update Learning
    epsilon = epsilon - Delta;
    
    % ------ Reset ---------
    
    %if(t==220)
    %   alfa_selected_EG = zeros(1,K); 
    %   Q_EG = zeros(K,1);
    %   % epsilon = 1;
    %   % Delta = 0.01;       
    %end
    
end

AccR_EG = cumsum(R_EG); % Note that this time, the minimum is the best

figure
yyaxis left
plot(ED1.*1E3,'k-');
hold
plot(ED2.*1E3,'b-');
hold
ylabel('Delay (ms)');

yyaxis right
plot(alfa_EG,'ro');

ylabel('Selected Action');
xlabel('Time');
legend('Delay Path 1','Delay Path 2','Action Selected');
axis([0 TimeHorizon 0 8])

figure
plot(Q_EG_disp(1,:));
hold
plot(Q_EG_disp(2,:));
plot(Q_EG_disp(3,:));
plot(Q_EG_disp(4,:));
plot(Q_EG_disp(5,:));
plot(Q_EG_disp(6,:));
plot(Q_EG_disp(7,:));
hold
legend('\alpha=0','\alpha=0.1','\alpha=0.25','\alpha=0.5','\alpha=0.75','\alpha=0.9','\alpha=1');
xlabel('Time');
ylabel('Q-table');


figure
yyaxis left
plot(AccR_EG);
ylabel('Accumulated Reward');

yyaxis right
plot(alfa_EG,'ro');

ylabel('Selected Action');
xlabel('Time');
axis([0 TimeHorizon 0 8])

figure
plot(w(1,:),'bx');
hold
plot(w(2,:),'ro');
hold
legend('Congestion Path 1','Congestion Path 2');
ylabel('Channel Occupancy');
xlabel('Time');

disp('Final Performance');
disp('Max delay (av) between path 1 and path 2 | Total Losses (sum of path 1 and path 2)');
disp([max([mean(ED1) mean(ED2)]) mean(ELosses1)+mean(ELosses2)]);


end




function [ED,ELosses]=PathDelay(K,B,L,C,w)

if(B > 0)
    ES = L/C;
    lambda=(B/L);
    a = lambda*ES+w;

    if a==1
        EN = K/2;
        Pb = 1/(K+1);
    else
        Pb = (1-a)*a^K/(1-a^(K+1));
        EN = a/(1-a) - ((K+1)*a^(K+1))/(1-a^(K+1));
    end
    ED = EN/(lambda*(1-Pb));
    ELosses = B*Pb;
else
    ED = 0;
    ELosses = 0;    
end

end